/* 
 * File:   Game.cpp
 *
 * The zone where the game takes place.  Holds all of the enemy objects.  Spawns
 * enemies from the negitive z wall.  Deletes enemies from the negitive z wall.
 */

#include "Game.h"
#include "glm/glm.hpp"
#include "State.h"
#include "FinalProject.h"
#include "GameOver.h"
#include "objects/Baseball.h"
#include "objects/Enemy.h"
#include "objects/Bird.h"
#include "objects/Plane.h"
#include "glm/gtx/simd_vec4.hpp"
#include "GlutTextHelper.h"
#include "objects/Cloud.h"

#define NUM_ENEMIES 2
#define PLANE_PERCENT 40
#define CLOUD_PERCENT 60
#define RAND ((float) rand() / RAND_MAX)
#define MIN_XY_VEL 0.25

using namespace std;

Game::Game()
{
    distance = 0;

    upperBounds = glm::vec3(40, 40, 10);
    lowerBounds = glm::vec3(-40, -40, -50);

    baseVelocity = 1.0;
    baseAcceleration = 0.0001;

    spawnChance = 0.05;
    spawnChanceInc = 0.0005;

    setCursorInvisible();
}

void Game::setCursorInvisible()
{
    vec3 ballPos = baseball.getPosition();
    glutWarpPointer(w2px(ballPos.x), w2py(ballPos.y));
    glutSetCursor(GLUT_CURSOR_NONE);
}

Game::Game(const Game& orig)
{
}

Game::~Game()
{
    glutSetCursor(GLUT_CURSOR_INHERIT);
}

void Game::draw()
{
    // draw baseball
    baseball.drawBaseball();

    // draw enemies
    for (set<Enemy*>::iterator iter = enemies.begin();
            iter != enemies.end();
            iter++)
    {
        (*iter)->drawEnemy();
    }

    std::string distStr = "Distance: ";
    char buff[50];
    sprintf(buff, "%ld", distance);
    drawText(distStr.append(buff).append("mi"), 0.002, vec3(0, 3.9, 0.51));

    if (isPaused())
    {
        std::string title = "Game Paused";
        drawText(title, 0.009, vec3(0, 0.15, 0.51));

        std::string subTitle = "Press \"P\" to Continue";
        drawText(subTitle, 0.004, vec3(0, -0.5, 0.51));

        std::string distStr = "Spawn Chance: ";
        char buff[50];
        sprintf(buff, "%ld", (long) round(spawnChance * 100));
        drawText(distStr.append(buff).append("%"), 0.0015, vec3(0, 3.68, 0.51));
    }
}

/**
 * Runs updateEnemy for every Enemy in the game.  Spawns or deletes enemies as
 * neccecary.
 */
void Game::update()
{
    bool isGameOver = false;

    if (!isPaused())
    {
        // spawn logic
        if (shouldSpawn())
        {
            enemies.insert(newEnemy());
        }

        // update logic
        baseball.updateBaseball();

        Enemy* curr;

        for (set<Enemy*>::iterator iter = enemies.begin();
                iter != enemies.end();
                iter++)
        {
            curr = (*iter);

            curr->updateEnemy();

            if (!curr->isCloud())
            {
                vec3 distanceVec = baseball.getPosition() - (*iter)->position;
                double dist = std::sqrt(std::pow(distanceVec.x, 2) +
                                        std::pow(distanceVec.y, 2) + std::pow(distanceVec.z, 2));

                if (dist < (*iter)->getRadius() + baseball.getRadius())
                {
                    //game over! Let this loop know that it needs to delete
                    //enemies now
                    isGameOver = true;
                }
            }

            // cleanup enemies that have gone past the baseball
            if (isGameOver || !isPastCamera(curr))
            {
                enemies.erase(curr);
                delete curr;
            }
        }

        // update baseVelocity
        baseVelocity += baseAcceleration;

        // update spawnChance
        spawnChance += spawnChanceInc;

        // update distance
        distance++;

        if (isGameOver)
            advanceState(new GameOver(distance));
    }
}

bool Game::isPastCamera(Enemy* enemy)
{
    return enemy->position.z <= upperBounds.z;
}

bool Game::shouldSpawn()
{
    return RAND < spawnChance;
}

Enemy* Game::newEnemy()
{
    vec3 spawnPos;
    spawnPos.x = (upperBounds.x - lowerBounds.x) * RAND + lowerBounds.x;
    spawnPos.y = (upperBounds.y - lowerBounds.y) * RAND + lowerBounds.y;
    spawnPos.z = lowerBounds.z;

    vec3 ballPos = baseball.getPosition();
    vec3 spawnVel = glm::normalize(ballPos - spawnPos);
    spawnVel.x = spawnVel.x * 1.5;
    spawnVel.y = spawnVel.y * 1.5;
    spawnVel.z = spawnVel.z * 0.5 * RAND + baseVelocity;

    // make sure nothing falls straight down
    if (spawnVel.x < MIN_XY_VEL)
    {
        spawnVel.x += 0.05 * glm::normalize(spawnVel.x);
    }

    if (spawnVel.y < MIN_XY_VEL)
    {
        spawnVel.y += 0.05 * glm::normalize(spawnVel.y);
    }

    // no acceleration
    vec3 spawnAcc = vec3(0);

    int determineEnemyType = ((float) rand() / RAND_MAX) * 100;
    
    if (determineEnemyType < PLANE_PERCENT)
    {
        return new Plane(spawnPos, spawnVel, spawnAcc);
    }
    if (determineEnemyType < CLOUD_PERCENT)
    {
        return new Cloud(spawnPos, spawnVel, spawnAcc);
    }
    else
    {
        return new Bird(spawnPos, spawnVel, spawnAcc);
    }
}

void Game::mouseMoved(int x, int y)
{
    if (!isPaused())
    {
        baseball.setPosition(glm::vec3(p2wx(x), p2wy(y), 0));
    }
}

void Game::pause()
{
    State::pause();

    glutSetCursor(GLUT_CURSOR_INHERIT);
}

void Game::resume()
{
    State::resume();

    setCursorInvisible();
}

void Game::printEnemies()
{
    for (set<Enemy*>::iterator iter = enemies.begin();
            iter != enemies.end();
            iter++)
    {
        cout << **iter << "\n";
    }

    cout << "\n";
}